import { highlightSpecialChars, drawSelection, highlightActiveLine, keymap, EditorView, ViewPlugin } from '@codemirror/view';
import { Annotation, Facet, combineConfig, StateField, EditorState, Compartment } from '@codemirror/state';
import { history, historyKeymap } from '@codemirror/history';
import { foldGutter, foldKeymap } from '@codemirror/fold';
import { indentOnInput, LanguageSupport } from '@codemirror/language';
import { lineNumbers } from '@codemirror/gutter';
import { defaultKeymap } from '@codemirror/commands';
import { bracketMatching } from '@codemirror/matchbrackets';
import { closeBrackets, closeBracketsKeymap } from '@codemirror/closebrackets';
import { highlightSelectionMatches, searchKeymap } from '@codemirror/search';
import { autocompletion, completionKeymap } from '@codemirror/autocomplete';
import { commentKeymap } from '@codemirror/comment';
import { rectangularSelection } from '@codemirror/rectangular-selection';
import { defaultHighlightStyle, HighlightStyle, tags } from '@codemirror/highlight';
import { lintKeymap, setDiagnostics } from '@codemirror/lint';
import React, { useEffect, forwardRef, useRef, useImperativeHandle } from 'react';
import { dequal } from 'dequal';
import { javascript } from '@codemirror/lang-javascript';
import { json } from '@codemirror/lang-json';
import { StandardSQL, schemaCompletion, keywordCompletion } from '@codemirror/lang-sql';

/**
 * @method define
 */
var formatDoc = Annotation.define();
Facet.define();
var formatConfig = Facet.define({
    combine: function combine(configs) {
        return combineConfig(configs, {
            parser: function (text) { return text; },
        }, {});
    },
});
var formatField_ = StateField.define({
    create: function create() { },
    update: function update(state, tr) {
        tr.state.facet(formatConfig);
        tr.annotation(formatDoc);
    },
    toJSON: function toJSON(value) { },
    fromJSON: function fromJSON(json) { },
});
function format(configs) {
    if ( configs === void 0 ) configs = {};

    return [formatField_, formatConfig.of(configs)];
}
function startFormat(view, configs) {
    var str = configs.parser(view.state.doc.toJSON().join("\n"));
    Promise.resolve(str).then(function (resolveStr) {
        view.dispatch({
            // annotations: formatDoc.of(),
            changes: [
                {
                    from: 0,
                    to: view.state.doc.length,
                    insert: resolveStr,
                } ],
        });
    });
}

/// This is an extension value that just pulls together a whole lot of
/// extensions that you might want in a basic editor. It is meant as a
/// convenient helper to quickly set up CodeMirror without installing
/// and importing a lot of packages.
///
/// Specifically, it includes...
///
///  - [the default command bindings](#commands.defaultKeymap)
///  - [line numbers](#gutter.lineNumbers)
///  - [special character highlighting](#view.highlightSpecialChars)
///  - [the undo history](#history.history)
///  - [a fold gutter](#fold.foldGutter)
///  - [custom selection drawing](#view.drawSelection)
///  - [multiple selections](#state.EditorState^allowMultipleSelections)
///  - [reindentation on input](#language.indentOnInput)
///  - [the default highlight style](#highlight.defaultHighlightStyle) (as fallback)
///  - [bracket matching](#matchbrackets.bracketMatching)
///  - [bracket closing](#closebrackets.closeBrackets)
///  - [autocompletion](#autocomplete.autocompletion)
///  - [rectangular selection](#rectangular-selection.rectangularSelection)
///  - [active line highlighting](#view.highlightActiveLine)
///  - [selection match highlighting](#search.highlightSelectionMatches)
///  - [search](#search.searchKeymap)
///  - [commenting](#comment.commentKeymap)
///  - [linting](#lint.lintKeymap)
///
/// (You'll probably want to add some language package to your setup
/// too.)
///
/// This package does not allow customization. The idea is that, once
/// you decide you want to configure your editor more precisely, you
/// take this package's source (which is just a bunch of imports and
/// an array literal), copy it into your own code, and adjust it as
/// desired.
var basicSetup = [
    lineNumbers(),
    highlightSpecialChars(),
    history(),
    foldGutter(),
    drawSelection(),
    EditorState.allowMultipleSelections.of(true),
    indentOnInput(),
    defaultHighlightStyle.fallback,
    bracketMatching(),
    closeBrackets(),
    autocompletion(),
    rectangularSelection(),
    highlightActiveLine(),
    highlightSelectionMatches(),
    format(),
    keymap.of(closeBracketsKeymap.concat( defaultKeymap,
        searchKeymap,
        historyKeymap,
        foldKeymap,
        commentKeymap,
        completionKeymap,
        lintKeymap )) ];

/**
 * @method useMount
 * @param  {()  => void} cb
 */
function useMount(cb) {
    useEffect(function () {
        cb && cb();
    }, []);
    return null;
}
/**
 * @method useUnmount
 * @param  {()    => void} cb
 */
function useUnmount(cb) {
    useEffect(function () {
        return function () {
            cb();
        };
    }, []);
    return null;
}

function sql(config) {
    if ( config === void 0 ) config = {};

    var configs = Object.assign({}, config);
    if (!configs.schema) {
        configs.schema = {};
    }
    var lang = configs.dialect || StandardSQL;
    return new LanguageSupport(lang.language, [
        schemaCompletion(configs),
        keywordCompletion(lang, !!configs.upperCaseKeywords) ]);
}

function viewChange(handler, editable) {
    if ( editable === void 0 ) editable = true;

    return EditorView.updateListener.of(function (ref) {
        var docChanged = ref.docChanged;
        var state = ref.state;

        if (docChanged) {
            handler &&
                editable &&
                handler(state.doc.toJSON().join(state.lineBreak));
        }
    });
}

// 定义minimapConfig facet
var minimapConfig = Facet.define({
    combine: function combine(configs) {
        var minimap;
        var explicit;
        var left;
        for (var c of configs) {
            minimap = c.minimap;
            explicit = c.explicit;
            left = c.left;
        }
        return { minimap: minimap, explicit: explicit, left: left };
    },
});
/**
 * @public Plugin
 *
 * 使用这个Plugin在Editor中生成minimap
 */
ViewPlugin.fromClass(/*@__PURE__*/(function () {
    function anonymous(view) {
        this.lineCache = new Map();
        view.state.facet(minimapConfig);
        this.render(view);
    }
    anonymous.prototype.update = function update (view) {
        if (!view.docChanged) {
            return;
        }
        var index = 0;
        // 遍历所有的line
        // FIXME 最好是能够只针对变化的行进行重新绘制，边里所有的行，实在是性能太差了
        while (index < view.state.doc.lines) {
            var line = view.state.doc.line(index + 1);
            this.lineCache.get(index);
            // console.log(view.view.visualLineAt(line.from))
            console.log(view.view.domAtPos(line.from));
            this.lineCache.set(index, line.text);
            index++;
        }
    };
    anonymous.prototype.render = function render (view) {
        if (!this.dom) {
            this.dom = document.createElement("canvas");
            this.dom.style.background = "blue";
            this.dom.style.height = "100%";
            this.dom.style.width = "120px";
            this.dom.style.position = "sticky";
            this.dom.style.right = "2px";
            this.dom.style.top = "0";
            this.dom.style.bottom = "0";
            this.dom.style.boxShadow = "red 0px 7px 5px";
            view.scrollDOM.appendChild(this.dom);
        }
    };
    // 给格子上色
    anonymous.prototype.syntax = function syntax (ref) {
        ref.from;
        ref.to;
        ref.color;
 };

    return anonymous;
}()), {});

// Using https://github.com/one-dark/vscode-one-dark-theme/ as reference for the colors
var chalky$1 = "#e5c07b", coral$1 = "#e06c75", cyan$1 = "#56b6c2", invalid$1 = "#ffffff", ivory$1 = "#abb2bf", stone$1 = "#7d8799", // Brightened compared to original to increase contrast
malibu$1 = "#61afef", sage$1 = "#98c379", whiskey$1 = "#d19a66", violet$1 = "#c678dd", darkBackground$1 = "#21252b", highlightBackground$1 = "#2c313a", background$1 = "#282c34", selection$1 = "#3E4451", cursor$1 = "#528bff";
/// The editor theme styles for One Dark.
var oneDarkTheme$1 = EditorView.theme({
    "&": {
        color: ivory$1,
        backgroundColor: background$1,
        "& ::selection": { backgroundColor: selection$1 },
        caretColor: cursor$1,
        height: '100%'
    },
    "&.cm-focused .cm-cursor": { borderLeftColor: cursor$1 },
    "&.cm-focused .cm-selectionBackground, .cm-selectionBackground": {
        backgroundColor: selection$1,
    },
    ".cm-panels": { backgroundColor: darkBackground$1, color: ivory$1 },
    ".cm-panels.cm-panels-top": { borderBottom: "2px solid black" },
    ".cm-panels.cm-panels-bottom": { borderTop: "2px solid black" },
    ".cm-searchMatch": {
        backgroundColor: "#72a1ff59",
        outline: "1px solid #457dff",
    },
    ".cm-searchMatch.cm-searchMatch-selected": {
        backgroundColor: "#6199ff2f",
    },
    ".cm-activeLine": { backgroundColor: highlightBackground$1 },
    ".cm-selectionMatch": { backgroundColor: "#aafe661a" },
    ".cm-matchingBracket, .cm-nonmatchingBracket": {
        backgroundColor: "#bad0f847",
        outline: "1px solid #515a6b",
    },
    ".cm-gutters": {
        backgroundColor: background$1,
        color: stone$1,
        border: "none",
    },
    ".cm-lineNumbers .cm-gutterElement": { color: "inherit" },
    ".cm-foldPlaceholder": {
        backgroundColor: "transparent",
        border: "none",
        color: "#ddd",
    },
    ".cm-tooltip": {
        border: "1px solid #181a1f",
        backgroundColor: darkBackground$1,
    },
    ".cm-tooltip-autocomplete": {
        "& > ul > li[aria-selected]": {
            backgroundColor: highlightBackground$1,
            color: ivory$1,
        },
        "& > ul > li > cm-completionIcon": {
            paddingRight: "2em",
        },
    },
}, { dark: true });
/// The highlighting style for code in the One Dark theme.
var oneDarkHighlightStyle$1 = HighlightStyle.define([
    { tag: tags.keyword, color: violet$1 },
    {
        tag: [
            tags.name,
            tags.deleted,
            tags.character,
            tags.propertyName,
            tags.macroName ],
        color: coral$1,
    },
    {
        tag: [tags.function(tags.variableName), tags.labelName],
        color: malibu$1,
    },
    {
        tag: [
            tags.color,
            tags.constant(tags.name),
            tags.standard(tags.name) ],
        color: whiskey$1,
    },
    { tag: [tags.definition(tags.name), tags.separator], color: ivory$1 },
    {
        tag: [
            tags.typeName,
            tags.className,
            tags.number,
            tags.changed,
            tags.annotation,
            tags.modifier,
            tags.self,
            tags.namespace ],
        color: chalky$1,
    },
    {
        tag: [
            tags.operator,
            tags.operatorKeyword,
            tags.url,
            tags.escape,
            tags.regexp,
            tags.link,
            tags.special(tags.string) ],
        color: cyan$1,
    },
    { tag: [tags.meta, tags.comment], color: stone$1 },
    { tag: tags.strong, fontWeight: "bold" },
    { tag: tags.emphasis, fontStyle: "italic" },
    { tag: tags.link, color: stone$1, textDecoration: "underline" },
    { tag: tags.heading, fontWeight: "bold", color: coral$1 },
    {
        tag: [tags.atom, tags.bool, tags.special(tags.variableName)],
        color: whiskey$1,
    },
    {
        tag: [tags.processingInstruction, tags.string, tags.inserted],
        color: sage$1,
    },
    { tag: tags.invalid, color: invalid$1 } ]);
/// Extension to enable the One Dark theme (both the editor theme and
/// the highlight style).
var oneDark$1 = [oneDarkTheme$1, oneDarkHighlightStyle$1];

// Using https://github.com/one-dark/vscode-one-dark-theme/ as reference for the colors
var chalky = "#e5c07b", coral = "#e06c75", cyan = "#56b6c2", invalid = "#ffffff", ivory = "#abb2bf", stone = "#7d8799", // Brightened compared to original to increase contrast
malibu = "#61afef", sage = "#98c379", whiskey = "#d19a66", violet = "#c678dd", darkBackground = "#21252b", highlightBackground = "#e2e2e2", background = "#ffffff", selection = "#3E4451", cursor = "#528bff";
/// The editor theme styles for One Dark.
var oneDarkTheme = EditorView.theme({
    "&": {
        color: ivory,
        backgroundColor: background,
        "& ::selection": { backgroundColor: selection },
        caretColor: cursor,
        height: '100%'
    },
    "&.cm-focused .cm-cursor": { borderLeftColor: cursor },
    "&.cm-focused .cm-selectionBackground, .cm-selectionBackground": {
        backgroundColor: selection,
    },
    ".cm-panels": { backgroundColor: darkBackground, color: ivory },
    ".cm-panels.cm-panels-top": { borderBottom: "2px solid black" },
    ".cm-panels.cm-panels-bottom": { borderTop: "2px solid black" },
    ".cm-searchMatch": {
        backgroundColor: "#72a1ff59",
        outline: "1px solid #457dff",
    },
    ".cm-searchMatch.cm-searchMatch-selected": {
        backgroundColor: "#6199ff2f",
    },
    ".cm-activeLine": { backgroundColor: highlightBackground },
    ".cm-selectionMatch": { backgroundColor: "#aafe661a" },
    ".cm-matchingBracket, .cm-nonmatchingBracket": {
        backgroundColor: "#bad0f847",
        outline: "1px solid #515a6b",
    },
    ".cm-gutters": {
        backgroundColor: background,
        color: stone,
        border: "none",
    },
    ".cm-lineNumbers .cm-gutterElement": { color: "inherit" },
    ".cm-foldPlaceholder": {
        backgroundColor: "transparent",
        border: "none",
        color: "#ddd",
    },
    ".cm-tooltip": {
        border: "1px solid #181a1f",
        backgroundColor: darkBackground,
    },
    ".cm-tooltip-autocomplete": {
        "& > ul > li[aria-selected]": {
            backgroundColor: highlightBackground,
            color: ivory,
        },
        "& > ul > li > cm-completionIcon": {
            paddingRight: "2em",
        },
    },
}, { dark: true });
/// The highlighting style for code in the One Dark theme.
var oneDarkHighlightStyle = HighlightStyle.define([
    { tag: tags.keyword, color: violet },
    {
        tag: [
            tags.name,
            tags.deleted,
            tags.character,
            tags.propertyName,
            tags.macroName ],
        color: coral,
    },
    {
        tag: [tags.function(tags.variableName), tags.labelName],
        color: malibu,
    },
    {
        tag: [
            tags.color,
            tags.constant(tags.name),
            tags.standard(tags.name) ],
        color: whiskey,
    },
    { tag: [tags.definition(tags.name), tags.separator], color: ivory },
    {
        tag: [
            tags.typeName,
            tags.className,
            tags.number,
            tags.changed,
            tags.annotation,
            tags.modifier,
            tags.self,
            tags.namespace ],
        color: chalky,
    },
    {
        tag: [
            tags.operator,
            tags.operatorKeyword,
            tags.url,
            tags.escape,
            tags.regexp,
            tags.link,
            tags.special(tags.string) ],
        color: cyan,
    },
    { tag: [tags.meta, tags.comment], color: stone },
    { tag: tags.strong, fontWeight: "bold" },
    { tag: tags.emphasis, fontStyle: "italic" },
    { tag: tags.link, color: stone, textDecoration: "underline" },
    { tag: tags.heading, fontWeight: "bold", color: coral },
    {
        tag: [tags.atom, tags.bool, tags.special(tags.variableName)],
        color: whiskey,
    },
    {
        tag: [tags.processingInstruction, tags.string, tags.inserted],
        color: sage,
    },
    { tag: tags.invalid, color: invalid } ]);
/// Extension to enable the One Dark theme (both the editor theme and
/// the highlight style).
var oneDark = [oneDarkTheme, oneDarkHighlightStyle];

var translateDiagnostics = function (input, view) {
    if ( input === void 0 ) input = {
    checkResult: { checkItems: [], rawSQL: "" },
};

    var checkItems = input.checkResult.checkItems; if ( checkItems === void 0 ) checkItems = [];
    var found = [];
    var maxLines = view.state.doc.lines;
    checkItems.forEach(function (item) {
        var error = item.error;
        var errorCode = item.errorCode;
        var lineOffsets = item.lineOffsets;
        lineOffsets.forEach(function (lineOffset) {
            var StartLine = lineOffset.StartLine;
            var EndLine = lineOffset.EndLine;
            var StartOffset = lineOffset.StartOffset;
            var EndOffset = lineOffset.EndOffset;
            var from, to;
            if (maxLines >= EndLine) {
                var ref = view.state.doc.line(StartLine + 1);
                var startFrom = ref.from;
                var startTo = ref.to;
                var ref$1 = view.state.doc.line(EndLine + 1);
                var endFrom = ref$1.from;
                var endTo = ref$1.to;
                if (startFrom === endFrom && startFrom === endTo) {
                    // same line
                    from = startFrom + StartOffset;
                    to = startTo + EndOffset;
                }
                else {
                    from = startFrom + StartOffset;
                    to = endFrom + EndOffset;
                }
                found.push({
                    from: from,
                    to: to,
                    severity: errorCode == -1 ? "error" : "warning",
                    message: error,
                });
            }
        });
    });
    return found;
};

/**
 * 对 codemirror6 的封装
 */
function objectWithoutProperties (obj, exclude) { var target = {}; for (var k in obj) if (Object.prototype.hasOwnProperty.call(obj, k) && exclude.indexOf(k) === -1) target[k] = obj[k]; return target; }
var LANGUAGE_EXTENSIONS = {
    javascript: javascript,
    sql: sql,
    json: json,
};
var THEMES = {
    dark: oneDark$1,
    light: oneDark,
};
function ReactCodemirror(props, ref) {
    var extensions = props.extensions; if ( extensions === void 0 ) extensions = [];
    var language = props.language;
    var langOptions = props.langOptions;
    var defaultValue = props.defaultValue;
    var onChange = props.onChange;
    var value = props.value;
    var theme = props.theme; if ( theme === void 0 ) theme = "dark";
    props.defaultLines;
    var editable = props.editable; if ( editable === void 0 ) editable = true;
    var diagnostics = props.diagnostics;
    var rest = objectWithoutProperties( props, ["extensions", "language", "langOptions", "defaultValue", "onChange", "value", "theme", "defaultLines", "editable", "diagnostics"] );
    var others = rest;
    var preLangOptions = useRef(langOptions);
    var element = useRef();
    var editor = useRef();
    var languageCompartment = useRef(new Compartment());
    var editableCompartment = useRef(new Compartment());
    var extensionsCompartment = useRef(new Compartment());
    var themeCompartment = useRef(new Compartment());
    useImperativeHandle(ref, function () { return ({
        format: function format(configs) {
            startFormat(editor.current, configs);
        },
    }); }, [editor]);
    useMount(function () {
        var languageEx = languageCompartment.current.of([
            // @ts-ignore
            LANGUAGE_EXTENSIONS[language](langOptions) ].filter(Boolean));
        var editoableEx = editableCompartment.current.of(EditorView.editable.of(editable));
        var extensionsEx = extensionsCompartment.current.of(extensions.filter(Boolean));
        var themeEx = themeCompartment.current.of(THEMES[theme]);
        var state = EditorState.create({
            doc: defaultValue,
            extensions: [
                basicSetup,
                languageEx,
                editoableEx,
                extensionsEx,
                themeEx,
                viewChange(onChange, editable) ],
        });
        editor.current = new EditorView({
            state: state,
            parent: element.current,
        });
    });
    // 1. language
    useEffect(function () {
        if (language === "sql" &&
            !dequal(langOptions, preLangOptions.current)) {
            editor.current.dispatch({
                effects: languageCompartment.current.reconfigure(
                // @ts-ignore
                LANGUAGE_EXTENSIONS[language](langOptions)),
            });
            preLangOptions.current = langOptions;
        }
    }, [langOptions, language]);
    // 2. extensions
    useEffect(function () {
        editor.current.dispatch({
            effects: extensionsCompartment.current.reconfigure([].concat( extensions ).filter(Boolean)),
        });
    }, [extensions]);
    // 3. theme, editable
    useEffect(function () {
        editor.current.dispatch({
            effects: themeCompartment.current.reconfigure([
                [THEMES[theme]].filter(Boolean),
                EditorView.editable.of(editable) ]),
        });
    }, [theme, editable]);
    var isSameTextAccordingToDoc = function (editor, value) {
        if ( value === void 0 ) value = "";

        var doc = editor.state.doc.toJSON();
        var nextDocText = value.split(editor.state.lineBreak);
        if (doc.length !== nextDocText.length) {
            return false;
        }
        var i = doc.length - 1;
        while (i >= 0) {
            if (doc[i] !== nextDocText[i]) {
                return false;
            }
            i--;
        }
        return true;
    };
    useEffect(function () {
        // 如果是静态的编辑器，那就把外部传入的value直接dispatch到EditorState中
        if (!isSameTextAccordingToDoc(editor.current, value)) {
            editor.current.dispatch({
                changes: [
                    {
                        from: 0,
                        to: editor.current.state.doc.length,
                        insert: value,
                    } ],
            });
        }
    }, [value]);
    /**
     * @description 监听变化的lint信息，同步到editor
     */
    useEffect(function () {
        if (!diagnostics) {
            return;
        }
        editor.current.dispatch(setDiagnostics(editor.current.state, translateDiagnostics(diagnostics, editor.current)));
    }, [diagnostics]);
    useUnmount(function () { return editor.current.destroy(); });
    return (React.createElement( 'div', Object.assign({}, { ref: element, className: "codemirror-editor-body" }, others)));
}
var ReactCodemirror$1 = forwardRef(ReactCodemirror);

export { ReactCodemirror$1 as ReactCodemirror };
